from __future__ import absolute_import

from willow.utils import deprecation

from .base import ImageBackend


class WandBackend(ImageBackend):
    def __init__(self, image):
        self.image = image

    def to_buffer(self):
        return 'RGB', self.image.size, self.image.make_blob('RGB')

# DOESNT WORK. SEE: https://github.com/dahlia/wand/issues/123
#    @classmethod
#    def from_buffer(cls, buf):
#        mode, size, data = buf
#        return cls(Image(blob=data, format=mode, width=size[0], height=size[1]))

    @classmethod
    def from_file(cls, f):
        wand_image = cls.get_wand_image()
        wand_api = cls.get_wand_api()

        f.seek(0)

        image = wand_image.Image(file=f)
        image.wand = wand_api.library.MagickCoalesceImages(image.wand)
        return cls(image)

    @classmethod
    def get_wand_image(cls):
        import wand.image
        return wand.image

    @classmethod
    def get_wand_api(cls):    
        import wand.api
        return wand.api

    @classmethod
    def check(cls):
        cls.get_wand_image()
        cls.get_wand_api()


@WandBackend.register_operation('get_size')
def get_size(backend):
    return backend.image.size


@WandBackend.register_operation('resize')
@deprecation.deprecated_resize_parameters
def resize(backend, size):
    backend.image.resize(size[0], size[1])


@WandBackend.register_operation('crop')
@deprecation.deprecated_crop_parameters
def crop(backend, rect):
    backend.image.crop(left=rect[0], top=rect[1], right=rect[2], bottom=rect[3])


@WandBackend.register_operation('save_as_jpeg')
def save_as_jpeg(backend, f, quality=85):
    with backend.image.convert('jpeg') as converted:
        converted.compression_quality = quality
        converted.save(file=f)


@WandBackend.register_operation('save_as_png')
def save_as_png(backend, f):
    with backend.image.convert('png') as converted:
        converted.save(file=f)


@WandBackend.register_operation('save_as_gif')
def save_as_gif(backend, f):
    with backend.image.convert('gif') as converted:
        converted.save(file=f)


@WandBackend.register_operation('has_alpha')
def has_alpha(backend):
    return backend.image.alpha_channel


@WandBackend.register_operation('has_animation')
def has_animation(backend):
    return backend.image.animation


@WandBackend.register_operation('get_wand_image')
def get_wand_image(backend):
    return backend.image.clone()



from functools import wraps
import warnings


class RemovedInWillow03Warning(DeprecationWarning):
    pass


def deprecated_resize_parameters(resize):
    @wraps(resize)
    def wrapper(backend, *args):
        if len(args) == 1:
            size = args[0]
        elif len(args) == 2:
            size = (args[0], args[1])

            warnings.warn(
                "Passing width and height to Image.resize() in separate arguments is now deprecated. "
                "Change image.resize(width, height) to image.resize((width,height))",
                RemovedInWillow03Warning)
        else:
            raise TypeError("resize() takes exactly 1 argument (%d given)" % len(args))

        return resize(backend, size)

    return wrapper


def deprecated_crop_parameters(crop):
    @wraps(crop)
    def wrapper(backend, *args):
        if len(args) == 1:
            rect = args[0]
        elif len(args) == 4:
            rect = (args[0], args[1], args[2], args[3])

            warnings.warn(
                "Passing left, top, right and bottom to Image.crop() in separate arguments is now deprecated. "
                "Change image.crop(left, top, right, bottom) to image.crop((left, top, right, bottom))",
                RemovedInWillow03Warning)
        else:
            raise TypeError("crop() takes exactly 1 argument (%d given)" % len(args))

        return crop(backend, rect)

    return wrapper



import unittest
import io
import imghdr

from willow.backends import pillow as pillow_backend


class TestPillowOperations(unittest.TestCase):
    def setUp(self):
        with open('tests/images/transparent.png', 'rb') as f:
            self.backend = pillow_backend.PillowBackend.from_file(f)

    def test_get_size(self):
        width, height = pillow_backend.get_size(self.backend)
        self.assertEqual(width, 200)
        self.assertEqual(height, 150)

    def test_resize(self):
        pillow_backend.resize(self.backend, 100, 75)
        self.assertEqual(self.backend.image.size, (100, 75))

    def test_crop(self):
        pillow_backend.crop(self.backend, 10, 10, 100, 100)
        self.assertEqual(self.backend.image.size, (90, 90))

    def test_save_as_jpeg(self):
        output = io.BytesIO()
        pillow_backend.save_as_jpeg(self.backend, output)
        output.seek(0)

        self.assertEqual(imghdr.what(output), 'jpeg')

    def test_save_as_png(self):
        output = io.BytesIO()
        pillow_backend.save_as_png(self.backend, output)
        output.seek(0)

        self.assertEqual(imghdr.what(output), 'png')

    def test_save_as_gif(self):
        output = io.BytesIO()
        pillow_backend.save_as_gif(self.backend, output)
        output.seek(0)

        self.assertEqual(imghdr.what(output), 'gif')

    def test_has_alpha(self):
        has_alpha = pillow_backend.has_alpha(self.backend)
        self.assertTrue(has_alpha)

    def test_has_animation(self):
        has_animation = pillow_backend.has_animation(self.backend)
        self.assertFalse(has_animation)

    def test_transparent_gif(self):
        with open('tests/images/transparent.gif', 'rb') as f:
            backend = pillow_backend.PillowBackend.from_file(f)

        self.assertTrue(pillow_backend.has_alpha(backend))
        self.assertFalse(pillow_backend.has_animation(backend))

        # Check that the alpha of pixel 1,1 is 0
        self.assertEqual(backend.image.convert('RGBA').getpixel((1, 1))[3], 0)

    def test_resize_transparent_gif(self):
        with open('tests/images/transparent.gif', 'rb') as f:
            backend = pillow_backend.PillowBackend.from_file(f)

        pillow_backend.resize(self.backend, 100, 75)

        self.assertTrue(pillow_backend.has_alpha(backend))
        self.assertFalse(pillow_backend.has_animation(backend))

        # Check that the alpha of pixel 1,1 is 0
        self.assertEqual(backend.image.convert('RGBA').getpixel((1, 1))[3], 0)

    def test_save_transparent_gif(self):
        with open('tests/images/transparent.gif', 'rb') as f:
            backend = pillow_backend.PillowBackend.from_file(f)

        # Save it into memory
        f = io.BytesIO()
        pillow_backend.save_as_gif(backend, f)

        # Reload it
        f.seek(0)
        backend = pillow_backend.PillowBackend.from_file(f)

        self.assertTrue(pillow_backend.has_alpha(backend))
        self.assertFalse(pillow_backend.has_animation(backend))

        # Check that the alpha of pixel 1,1 is 0
        self.assertEqual(backend.image.convert('RGBA').getpixel((1, 1))[3], 0)

    @unittest.expectedFailure # Pillow doesn't support animation
    def test_animated_gif(self):
        with open('tests/images/newtons_cradle.gif', 'rb') as f:
            backend = pillow_backend.PillowBackend.from_file(f)

        self.assertFalse(pillow_backend.has_alpha(backend))
        self.assertTrue(pillow_backend.has_animation(backend))

    @unittest.expectedFailure # Pillow doesn't support animation
    def test_resize_animated_gif(self):
        with open('tests/images/newtons_cradle.gif', 'rb') as f:
            backend = pillow_backend.PillowBackend.from_file(f)

        pillow_backend.resize(self.backend, 100, 75)

        self.assertFalse(pillow_backend.has_alpha(backend))
        self.assertTrue(pillow_backend.has_animation(backend))
